@@ -62,8 +62,7 @@ def get_procentile ( vals, procentile = 50 ):
 
 pi0 = lambda pvs : 1.
 
-def padded_rolling_window ( tv, tau ) :
-        # SIMPLY RETURN THE PADDED INDICES
+def padded_rolling_window ( tv, tau ) : # SIMPLY RETURN THE PADDED INDICES
 	if tau==1 :
 		return ( tv )
 	if len(tv)<tau :
@@ -106,6 +105,204 @@ def svd_reduced_mean ( x,axis=0,keep=[0] ) :
                 return ( xred )
     return ( x )
 
+import math
+# YOU SHOULD READ Numerical Analysis by Burden and Faires
+def rich_rot ( a , b , direction = 0 ) :
+    if a==0 and b==0 :
+        c = 0
+        s = 0
+        r = 0
+    else :
+        r = math.sqrt( a*a + b*b )
+        if direction == 0 :
+            if a == 0 :
+                s = r / b
+                c = 0
+            else :
+                s = b / r
+                c = ( r - s*b ) / a
+        else :
+            if a == 0 :
+                s = 0
+                c = r / b
+            else :
+                s = - a / r
+                c = ( r - s*b ) / a
+    return ( c , s , r )
+
+def diagonalize_2b2( A , tol=1E-13 , maxiter = 10 ) :
+    M   = A[:2,:2].copy()
+    M0  = A[:2,:2].copy()
+    k   = 0
+    ERR = 1
+    G_  = None
+    H_  = None
+    for k in range( maxiter ) :
+        # LEFT
+        c,s,r = rich_rot( M0[0,0],M0[1,0])
+        G0    = np.array( [[c,s],[-s,c]] )
+        M     = np.dot( G0 , M0 )
+        # RIGHT
+        M = M.T
+        c,s,r = rich_rot( M[0,0],M[1,0])
+        H0    = np.array( [[c,s],[-s,c]] )
+        M     = np.dot( H0 , M )
+        # BUILD
+        M0 = M.T
+        ERR = np.sqrt( M0[1,0]**2+M0[0,1]**2 )
+        if G_ is None :
+            G_ = G0
+        else :
+            G_ = np.dot(G0,G_)
+        if H_ is None :
+            H_ = H0
+        else :
+            H_ = np.dot(H0,H_)
+        if ERR < tol :
+            break
+    return ( G_ , M0 , H_ )
+
+def skew_zero ( shape ):
+    return ( np.zeros(np.prod(shape)).reshape(shape) )
+
+def skew_eye ( shape ) :
+    Z = skew_zero(shape)
+    n = np.min(shape)
+    for i in range(n):
+        Z[i][i] = 1
+    return ( Z )
+
+def kth_householder ( A , k ):
+    # THE K:TH HOUSHOLDER ITERATION
+    # NOMECLATURE CORRESPONDS TO GOLUBS PAPER
+    A  = np .array( A )
+    n_ , m_ = np .shape(A)
+    if n_ < 2 :
+        return ( A )
+    k0 = k
+    k1 = k+1
+
+    alpha = ( 2*(A[k1][k0]<0)-1 )
+    alpha = alpha * np.sqrt( sum([ a**2 for a in A.T[k0][k1:] ]) )
+    r  = np.sqrt ( 0.5*(alpha**2-A[k1][k0]*alpha) )
+    v_ = [ 0 for z in range(k1) ] ; v_ .append( (A[k1][k0]-alpha)*0.5/r )
+    [ v_ .append ( (0.5/r) * A[j][k0] ) for j in range(k1+1,n_) ]
+    v_ = np.array( v_ )
+    Pk = np.eye( n_ ) - 2*np.array( [ v*w for v in v_ for w in v_ ] ).reshape(n_,n_)
+    Qk = Pk
+
+    if n_ != m_ :
+        alpha = ( 2*(A[k0][k1]<0)-1 )
+        alpha = alpha * np.sqrt( sum([ a**2 for a in A[k0][k1:] ]) )
+        r  = np.sqrt ( 0.5*(alpha**2-A[k0][k1]*alpha) )
+        w_ = [ 0 for z in range(k1) ] ; w_ .append( (A[k0][k1]-alpha)*0.5/r )
+        [ w_ .append ( (0.5/r) * A[k0][j] ) for j in range(k1+1,m_) ]
+        w_ = np.array( w_ )
+        Qk = np.eye( m_ ) - 2*np.array( [ v*w for v in w_ for w in w_ ] ).reshape(m_,m_)
+
+    Ak = np.dot( np.dot( Pk,A ),Qk )
+    return ( Pk,Ak,Qk )
+
+def Householder_transformation ( A ):
+    A     = np.array( A )
+    n = np.min( np.shape( A ) )
+    if n < 2 :
+        return ( A )
+    P = []
+    P0 , A0 , Q0 = kth_householder( A,k=0 )
+    P .append([P0,A0,Q0])
+    for k in range( 1 , n-1 ) : # ends at n-2
+        P1 , A1 , Q1 = kth_householder( A0,k=k )
+        A0 = A1
+        P  .append([P1,A1,Q1])
+    return ( A1 , P )
+
+def Householder_reduction ( A ):
+    A     = np.array( A )
+    n = np.min( np.shape( A ) )
+    if n < 2 :
+        return ( A )
+    P0 , A0 , Q0 = kth_householder( A,k=0 )
+    for k in range( 1 , n-1 ) : # ends at n-2
+        P1 , A1 , Q1 = kth_householder( A0,k=k )
+        A0 = A1
+        P0 = np.dot( P0 , P1 )
+        Q0 = np.dot( Q0 , Q1 )
+    U  = P0
+    S  = A1
+    VT = Q0.T
+    return ( U , S , VT )
+
+def diagonalize_tridiagonal ( tridiagonal ,
+            maxiter = 1000 ,
+            tol     = 1E-16 ):
+
+        quench = lambda A_,tol_ : np.array([[ a if a**2>tol_**2 else 0 for a in al] for al in A_])
+        dot_assign = lambda  A_, B_ : B_ if A_ is None else np.dot( B_ , A_ )
+
+        S = tridiagonal.copy()
+        n_ , m_ = np.shape( S )
+
+        sI = skew_eye ( [ n_ , n_ ] )
+        tI = skew_eye ( [ m_ , m_ ] )
+        zI = skew_eye ( np.shape(S) )
+        GI = None
+        HI = None
+        #
+        sI_   = sI.copy()
+        tI_   = tI.copy()
+        nm_   = np.min(np.shape(S)) - 1
+        #
+        for k in range ( maxiter ) :
+            for i in range ( nm_ ):
+                sI_   = sI.copy()
+                tI_   = tI.copy()
+                if True :
+                    A     = S[ i:i+2 , i:i+2 ].copy()
+                    if True :
+                        G , Z , H = diagonalize_2b2 ( A , tol=tol , maxiter=maxiter )
+                        sI_[ i:i+2 , i:i+2 ] = G
+                        GI = dot_assign(GI,sI_)
+
+                        tI_[ i:i+2 , i:i+2 ] = H
+                        HI = dot_assign(HI,tI_)
+
+                        S =  np.dot( np.dot( sI_ , S ) , tI_.T )
+                        for ir in range( 2,nm_+1-i ):
+                            ii  = i
+                            jj  = i+ir
+                            idx = [ (ii,ii),(ii,jj),(jj,ii),(jj,jj) ]
+                            jdx = [ (0,0),(0,1),(1,0),(1,1) ]
+                            A   = np.array( [ S[i] for i in idx] ).reshape(2,2) # ROW ORDERED
+                            G , Z , H = diagonalize_2b2 ( A , tol=tol , maxiter=maxiter )
+                            sI_   = sI.copy()
+                            tI_   = tI.copy()
+                            H = H.T
+                            for i_,j_ in zip(idx,jdx) :
+                                sI_[i_] = G[j_]
+                                tI_[i_] = H[j_]
+                            tI_= tI_.T
+                            GI = dot_assign(GI,sI_)
+                            HI = dot_assign(HI,tI_)
+                            S =  np.dot( np.dot( sI_ , S ) , tI_.T )
+            ERR = sum( np.diag(S[:nm_],-1)**2 ) + sum( np.diag(S[:nm_] ,1)**2 )
+            if ERR < 2.0*tol**2 :
+                break;
+        return ( GI.T , S , HI )
+
+def AugumentedReducedDecomposition ( A , maxiter=1000 , tol=1E-16 ):
+    P , Z , QT = Householder_reduction( A )
+    G , S , HT = diagonalize_tridiagonal( Z , maxiter=maxiter , tol=tol )
+    U  = np.dot(P,G)
+    VT = np.dot(HT,QT)
+    return ( U,S,VT )
+
+def AugumentedSingularDecomposition( A , maxiter=1000 , tol=1E-16 ):
+    return ( AugumentedReducedDecomposition ( A,  maxiter=maxiter , tol=tol ) )
+
+def ASVD( A , maxiter=1000 , tol=1E-16 ):
+    return ( AugumentedReducedDecomposition ( A,  maxiter=maxiter , tol=tol ) )
+
 from sklearn.decomposition import PCA
 dimred = PCA ( n_components = 1 )
 
